/**
 * This is where all of the constants I'd like to use throughout the 
 *   ray-tracer will live
 * 
 * @author	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date	09.11.11
 */
#ifndef CONSTANTS_H
#define CONSTANTS_H


#include "trax.hpp"			// for min/max definitions


/// raytracer constants namespace
namespace MYRTC {

	static const float GLOBAL_EPS = 1.e-7f;					/// global epsilon
	static const float GLOBAL_EPS_SQ = GLOBAL_EPS*GLOBAL_EPS;

	/**
	 * Some Camera FOV constants so we don't have to re-compute 
	 *   them all the time
	 */
	static const float CAM_FOV_90_DEG = 1.0f;				/// 90 deg fov
	static const float CAM_FOV_60_DEG = 0.577350269f;		/// 60 deg fov
	static const float CAM_FOV_45_DEG = 0.414213562f;		/// 45 deg fov
	static const float CAM_FOV_22_DEG = 0.194380309f;		/// 22 deg fov
	static const float CAM_FOV_193_DEG= 0.35f;				/// 19.3 deg fov


	/**
	 * Ray intersection point constants
	 */
	static const float RAY_EPS = 1.e-6f;					/// closest hit point for a ray
	static const float RAY_INF = 10000000.f;				/// furthest hit point for a ray
	static const float RAY_INTER_STEP = 1.e-5f;				/// step for a ray after a hit point


	/**
	 * Material properties that we can use for reflections/refractions
	 */
	static const float MATETA_AIR			= 1.0003f;
	static const float MATETA_WATER			= 1.33f;
	static const float MATETA_ACRYLIC		= 1.49f;
	static const float MATETA_CROWN_GLASS	= 1.52f;
	static const float MATETA_DIAMOND		= 2.42f;


	/**
	 * Types of objects we have implemented -> this is intended
	 *   to be used by our inheritance-less code... and casting
	 *   void pointers :/
	 */
	typedef enum {
		OT_UNKNOWN = 0,			/// unknown object type
		OT_SPHERE,				/// sphere object type
		OT_TRIANGLE,			/// triangle object type
		OT_TRIANGLE_TRAX,		/// triangle object type, BUT must be loaded from TRAX!
		OT_BOX,					/// box object type
	} OBJ_TYPE;


	/**
	 * Clamping function to [0,1] range
	 */
	static inline const float clamp01( const float &v ) {
		return max( min( v, 1.f ), 0.f );
	}

	/**
	 * Clamping function to [A,B] range
	 */
	static inline const float clampAB( const float &v, const float &A, const float &B ) {
		return max( min( v, B ), A );
	}

	/**
	 * Swaps two values by messing with addresses
	 * @param a - a value
	 * @param b - b value
	 */
	static inline const void swap( float *aPtr, float *bPtr ) {
		float * const tmpPtr = aPtr;
		aPtr = bPtr;
		bPtr = tmpPtr;
	}
};

#endif // CONSTANTS_H
