/**
 * Simple class implementing a box to be used by the BVH code
 *
 * @author	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date	09.25.11
 */
#ifndef GP_BOX_H
#define GP_BOX_H

#include "trax.hpp"
#include "Vector.h"
#include "RayPrimitive.h"


/// basic box implementation
class gpBox {
	public:
		static const int SIZE_IN_TRAX = 8;			/// number of 4Bs (aka sizeof(float)) this box takes up

		/**
		 * Constructor to create an empy 2x2x2 box centered at (0,0,0)
		 * WARNING: for speed - no checking on bounds!
		 * 
		 * @param minVals - lower bottom left corner of the box
		 * @param maxVals - upper top right corner of the box
		 */
		gpBox( const Vector &minVals=Vector(-1.f,-1.f,-1.f), const Vector &maxVals=Vector(1.f,1.f,1.f) ) {
			pts[0].set( minVals );
			pts[1].set( maxVals );
		}


		/**
		 * Reads this box from TRaX memory
		 * @param addr - start address of this triangle
		 */
		inline void set( const int &addr ) {

			// read bounds
			pts[0].set( addr );
			pts[1].set( addr+3 );

			// read number of children and starting triangle address or id of BVH child
			// without this -> saves us some memory load time -> faster!
		}


		/**
		 * Intersects a ray with this box and returns the near/far intersection points
		 * NOTE: if one intersection is behind the ray.minT(), t0out is set to ray.minT()
		 *
		 * @param t0out - out nearest ray parameter for intersection (or ray.minT)
		 * @param t1out - out the 2nd ray parameter (either farthest or behind)
		 * @param ray	- ray to test against
		 */
		bool intersectDetailed( float &t0out, float &t1out, const RayPrimitive &ray ) const;



	// public for speed
		Vector pts[2];			/// min/max points
};


#endif // GP_BOX_H
