/**
 * Implementation of the sphere primitive
 *
 * @author:	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date:	09.05.11
 */
#include "gpSphere.h"
#include "IntersectionRecord.h"


/// Get normal based on the hit location
void gpSphere::getNormal( Vector &hitNorm, const Vector &hitLoc ) const {
	hitNorm.set( hitLoc );
	hitNorm.sub( m_vCenter );
	//hitNorm.normalize();
	hitNorm.mult( m_fRadInv );	// this gets about 3-4 decimal places... which is fine for our scene
}


/// Intersection between a ray and this primitive - (use for primary rays!)
bool gpSphere::intersectDetailed( RayPrimitive &ray, IntersectionRecord &rec ) const {

	// shift ray into sphere's coordinate frame (aka sphere is at (0,0,0))
	Vector rayOrgShifted( ray.m_vOrigin );
	rayOrgShifted.sub( m_vCenter );

	// get the quadratic equation terms -> we assumed ray is normalized!
	const float b = ray.m_vDirection.dot( rayOrgShifted );
	const float c = rayOrgShifted.lengthSq() - m_fRadSq;

	// check if discriminant is negative -> no intersection
	float disc    = b*b - c;
	if( disc <= 0.f )
		return false;
	disc = sqrt( disc );

	// compute intersections and order them from smallest to largest
	float t0 = -b + disc;
	float t1 = -b - disc;
	if( t0 > t1 ) {
		const float tmp = t0;
		t0 = t1;
		t1 = tmp;
	}

	// test against ray parameters
	if( t0 > ray.m_fMaxT || t1 < ray.m_fMinT )
		return false;
	float tHit = t0;
	if( t0 < ray.m_fMinT ) {
		tHit = t1;
		if( tHit > ray.m_fMaxT )
			return false;
	}

	// update intersection record
	rec.m_fHitRayT	= tHit;
#ifdef DO_MULTIOBJECT
	rec.m_pGeomHit	= (void*)this;
	rec.m_oType		= MYRTC::OT_SPHERE;
	rec.m_pGeomMat	= m_pMat;
#endif
	ray.m_fMaxT		= tHit;
	return true;
}


/// Intersection between a ray and this primitive - (use for shadow rays!)
bool gpSphere::intersectQuick( const RayPrimitive &ray ) const {

	// shift ray into sphere's coordinate frame (aka sphere is at (0,0,0))
	Vector rayOrgShifted( ray.m_vOrigin );
	rayOrgShifted.sub( m_vCenter );

	// get the quadratic equation terms
	const float a = ray.m_vDirection.lengthSq();
	const float b = 2.0f*( ray.m_vDirection.dot( rayOrgShifted ) );
	const float c = rayOrgShifted.lengthSq() - m_fRadSq;

	// check if discriminant is negative -> no intersection
	float disc    = b*b - 4.f * a * c;
	if( disc <= 0.f )
		return true;
	disc = sqrt( disc );

	// compute intersections and order them from smallest to largest
	float q;
	if( b < 0 )
		q = (-0.5f) * (b - disc);
	else
		q = (-0.5f) * (b + disc);
	float t0 = q / a;
	float t1 = c / q;
	if( t0 > t1 ) {
		const float tmp = t0;
		t0 = t1;
		t1 = tmp;
	}

	// test against ray parameters
	if( t0 > ray.m_fMaxT || t1 < ray.m_fMinT )
		return true;
	if( t0 < ray.m_fMinT || t1 > ray.m_fMaxT )
		return true;

	// we're ok :D
	return false;
}


/// Intersection between a ray and this primitive - (use for primary rays!)
bool gpSphere::notIntersectQuickNormed( const RayPrimitive &ray ) const {

	// shift ray into sphere's coordinate frame (aka sphere is at (0,0,0))
	Vector rayOrgShifted( ray.m_vOrigin );
	rayOrgShifted.sub( m_vCenter );

	// get the quadratic equation terms -> we assumed ray is normalized!
	const float b = ray.m_vDirection.dot( rayOrgShifted );
	const float c = rayOrgShifted.lengthSq() - m_fRadSq;

	// check if discriminant is negative -> no intersection
	float disc    = b*b - c;
	if( disc <= 0.f )
		return true;
	disc = sqrt( disc );

	// compute intersections and order them from smallest to largest
	float t0 = -b + disc;
	float t1 = -b - disc;
	if( t0 > t1 ) {
		const float tmp = t0;
		t0 = t1;
		t1 = tmp;
	}

	// test against ray parameters
	if( t0 > ray.m_fMaxT || t1 < ray.m_fMinT )
		return true;
	if( t0 < ray.m_fMinT || t1 > ray.m_fMaxT )
		return true;

	// we're ok :D
	return false;
}
