/**
 * Implementation of the simple triangle class. This takes care of
 *   the intersection code
 *
 * @author	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date	09.12.11
 */

#include "Constants.h"
#include "gpTriangle.h"
#include "IntersectionRecord.h"


/// Intersection between a ray and a this triangle
bool gpTriangle::intersectDetailed( RayPrimitive &ray, IntersectionRecord &rec ) const {

	Vector edge1( m_vPoint0 );
	Vector edge2( m_vPoint2 );

	// are we dealing with points?
#if TRIANGLES_STORE_EDGES==0
	edge1.sub( m_vPoint1 );
	edge2.sub( m_vPoint1 );
#endif

	// test if triangle isn't too skinny
	Vector r1;
	ray.m_vDirection.cross( r1, edge2 );		// r1 = V x e2
	const float denom = r1.dot( edge1 );
	if( denom*denom < MYRTC::GLOBAL_EPS_SQ )
		return false;

	// get the 1st barycentric coordinate and test it
	const float invDenom = 1.f / denom;
	Vector s( ray.m_vOrigin );
	s.sub( m_vPoint1 );
	const float b1 = s.dot( r1 ) * invDenom;
	if( b1<0.f || b1>1.f )
		return false;

	// get the 2nd barycentric coordinate and test it
	Vector r2;
	s.cross( r2, edge1 );						// r2 = s x e1
	const float b2 = r2.dot( ray.m_vDirection ) * invDenom;
	if( b2<0.f || b1+b2>1.f )
		return false;

	// compute ray parameter t and test it
	const float t = r2.dot( edge2 ) * invDenom;
	if( t > ray.m_fMaxT || t < ray.m_fMinT )
		return false;


	// compute all intersection record stuff
	rec.m_fHitRayT		= t;
#ifdef DO_MULTIOBJECT
	rec.m_pGeomHit		= (void*)this;
	rec.m_oType			= MYRTC::OT_TRIANGLE;
	rec.m_pGeomMat		= m_pMat;
#endif
	rec.m_iGeomHitAddr	= m_objAddr;
	ray.m_fMaxT			= t;
	return true;
}


/// Intersection between a ray and this triangle
bool gpTriangle::intersectQuick( const RayPrimitive &ray ) const {

	Vector edge1( m_vPoint0 );
	Vector edge2( m_vPoint2 );

	// are we dealing with points?
#if TRIANGLES_STORE_EDGES==0
	edge1.sub( m_vPoint1 );
	edge2.sub( m_vPoint1 );
#endif

	// test if triangle isn't too skinny
	Vector r1;
	ray.m_vDirection.cross( r1, edge2 );		// r1 = V x e2
	const float denom = r1.dot( edge1 );
	if( denom*denom < MYRTC::GLOBAL_EPS_SQ )
		return true;

	// get the 1st barycentric coordinate and test it
	const float invDenom = 1.f / denom;
	Vector s( ray.m_vOrigin );
	s.sub( m_vPoint1 );
	const float b1 = s.dot( r1 ) * invDenom;
	if( b1<0.f || b1>1.f )
		return true;

	// get the 2nd barycentric coordinate and test it
	Vector r2;
	s.cross( r2, edge1 );						// r2 = s x e1
	const float b2 = r2.dot( ray.m_vDirection ) * invDenom;
	if( b2<0.f || b1+b2>1.f )
		return true;

	// compute ray parameter t and test it
	const float t = r2.dot( edge2 ) * invDenom;
	if( t > ray.m_fMaxT || t < ray.m_fMinT )
		return true;

	// survived all tests -> intersection!
	return false;
}
