#include "BVH.h"
#include "Triangle.h"
#include <float.h>
#include <assert.h>
#include <limits.h>
#include <queue>

#include <algorithm>

//#define TREE_ROTATIONS

const float BVH_C_isec = 10.f;
const float BVH_C_trav = 10.f;

class BVHNode {
public:
  BVHNode(){}
  float box_min[3];
  float box_max[3];
  int num_children; // -1 for interior nodes
  int start_child;  // direct block pointer for list of triangles

  int total_sub_nodes;
  float cost;

  void makeLeaf(int first_child, int count) {
#ifdef TREE_ROTATIONS
    assert(count == 1);
#endif
    start_child = first_child;
    num_children = count;
  }

  void makeInternal(int first_child) {
    start_child = first_child;
    num_children = -1;
  }

  bool isLeaf() const {
    return num_children != -1;
  }

  void LoadIntoMemory(int &memory_position,
                      int max_memory,
                      FourByte* memory) {
    // write myself into memory
    for (int i = 0; i < 3; i++) {
      memory[memory_position].fvalue = box_min[i];
      memory_position++;
    }
    for (int i = 0; i < 3; i++) {
      box_max[i] += 0.001;  // don't allow for 2d boxes
      memory[memory_position].fvalue = box_max[i];
      memory_position++;
    }
    memory[memory_position].ivalue = num_children;
    memory_position++;

    memory[memory_position].ivalue = start_child;
    memory_position++;
  }

};

BVH::~BVH()
{
  delete[] nodes;
}

BVH::BVH(std::vector<Triangle*>* _triangles, bool duplicate) {
  duplicate_BVH = duplicate;
  triangles = _triangles;
  num_tris = triangles->size();
  nodes = new BVHNode[2*triangles->size()];
  if(!nodes)
    {
      perror("in BVH::BVH: out of memory\n");
      exit(1);
    }
  int nextFree = 1;
  num_nodes = 0;
  build(0, 0, triangles->size(), nextFree, 0);
  updateBounds(0);
  computeSubtreeSize(0);
  oldComputeNodeCost(0);
}



void BVH::LoadIntoMemory(int &memory_position,
                         int max_memory,
                         FourByte* memory) {
  start_secondary_tris = -1;
  start_secondary_nodes = -1;
  printf("BVH bounds [%f %f %f] [%f %f %f]\n", 
	 nodes[0].box_min[0], nodes[0].box_min[1], nodes[0].box_min[2], 
	 nodes[0].box_max[0], nodes[0].box_max[1], nodes[0].box_max[2]);


  // align the BVH to cache line boundaries (root node starts mid-line, so that children are cache line-aligned)
  // want root node to be 8-word aligned and first child to be 16-word aligned
  memory_position = memory_position + (8 - (memory_position % 8));
  // load the first copy
  start_nodes = memory_position;
  LoadNodes(memory_position, max_memory, memory, start_nodes);
  start_tris = memory_position;
  LoadTriangles(memory_position, max_memory, memory);

  // load the second copy (if necessary)
  if(duplicate_BVH)
    {
      memory_position = memory_position + (8 - (memory_position % 8)); // align to cache-line boundary
      start_secondary_nodes = memory_position;
      LoadNodes(memory_position, max_memory, memory, start_secondary_nodes);
      start_secondary_tris = memory_position;
      LoadTriangles(memory_position, max_memory, memory);
    }

  // Keep everything else separate from the actual BVH so that nodes remain 8-word aligned

  // now load the SAH costs for tree rotations
  start_costs = memory_position;
  for(int i = 0; i < num_nodes; i++)
    {
      if(nodes[i].num_children < 0)
	memory[memory_position++].fvalue = 0.0f;
      else
	memory[memory_position++].fvalue = nodes[i].num_children * BVH_C_isec;
    }

  // now load the subtree sizes
  start_subtree_sizes = memory_position;
  for(int i = 0; i < num_nodes; i++)
    {
      memory[memory_position++].ivalue = nodes[i].total_sub_nodes;
    }

  // now load the rotated flags
  // Note: this could be compressed down to 1 bit per node, but who cares
  //       It could even be hidded in one of the other node data fields, 
  //       like num_children
  start_rotated_flags = memory_position;
  for(int i = 0; i < num_nodes; i++)
    memory[memory_position++].ivalue = 0;


}

void BVH::LoadNodes(int &memory_position,
	       int max_memory,
		    FourByte* memory, int start_node_copy)
{
  for (int i = 0; i < num_nodes; i++) 
    {
      nodes[i].LoadIntoMemory(memory_position, max_memory, memory);
    }
  // after we've loaded all the nodes, go through and fix up the
  // memory references (in memory.. ugh)
  int triangle_base_addr = memory_position;
  for (int i = 0; i < num_nodes; i++) 
    {
      if (nodes[i].isLeaf()) 
	{
	  // determine fixup and place in memory
	  int real_addr = triangle_base_addr + 11 * nodes[i].start_child; 
	  int node_addr = start_node_copy + i * 8;
	  memory[node_addr + 7].ivalue = real_addr; 
	}
    }
}
void BVH::LoadTriangles(int &memory_position,
		   int max_memory,
		   FourByte* memory)
{
  for (size_t i = 0; i < inorder_tris.size(); i++) 
    {
      inorder_tris[i]->object_id = i;
      inorder_tris[i]->LoadIntoMemory(memory_position, max_memory, memory);
    }
}

// Walks the tree and computes the number of children below 
// (including the node) each subtree
int BVH::computeSubtreeSize(int node_id)
{
  BVHNode *n = &nodes[node_id];
  if(n->num_children > 0) // leaf node
    {
      n->total_sub_nodes = 1;
      return 1;
    }
  n->total_sub_nodes = 1; // to account for myself
  n->total_sub_nodes += computeSubtreeSize(n->start_child); // left child
  n->total_sub_nodes += computeSubtreeSize(n->start_child + 1); // right child
  return n->total_sub_nodes;
}

// Rearranges nodes in memory so that children always come after their parent
// Necessary for the simple traversal technique the RT uses for refitting and rotating
void BVH::reorderNodes(int start_scene, FourByte *memory)
{
  BVHNode *unordered = new BVHNode[num_nodes];
  // copy the nodes out of memory since we're going to write over them
  for(int i=0; i < num_nodes; i++)
    unordered[i] = loadNode(i, start_scene, memory);
  
  int memory_position = start_scene;
  int loaded_so_far = 0;
  // breadth first traversal of the tree and write the nodes in the correct order
  std::queue<int> q;
  q.push(0);
  while(!q.empty())
    {
      BVHNode temp = unordered[q.front()];
      if(temp.num_children < 0)
	{	
	  q.push(temp.start_child);
	  q.push(temp.start_child + 1);	  	  
	  temp.start_child = q.size() - 2 + loaded_so_far;  
	}
      temp.LoadIntoMemory(memory_position, 0, memory); 
      loaded_so_far++;
      q.pop();
    }
  delete[] unordered;
}

float BoxArea(const float box_min[3], const float box_max[3]) {
  float dx = box_max[0] - box_min[0];
  float dy = box_max[1] - box_min[1];
  float dz = box_max[2] - box_min[2];

  return 2.f * (dx * dy + dx * dz + dy * dz);
}

// dot -Tgif initialBVH.dot -o out.gif
void BVH::writeDOT(const char *filename, int start_scene, FourByte *memory, int verbosity, int maxDepth)
{
  FILE *output = fopen(filename, "w");
  if(output == NULL)
    {
      printf("Error, could not open DOT file for writing: %s\n", filename);
      exit(1);
    }
  BVHNode root = loadNode(0, start_scene, memory);
  fprintf(output, "graph BVH {\n");
  //fprintf(output, "Node0 [label=\"0:[%f %f %f]-[%f %f %f](%f),1, I\"];\n", 
  //  root.box_min[0], root.box_min[1], root.box_min[2],
  //  root.box_max[0], root.box_max[1], root.box_max[2],
  //  memory[start_costs].fvalue);
  unsigned char red = 127;
  unsigned char green = 127;
  if(verbosity)
    fprintf(output, "Node0 [label=\"0, %d\"];\n", nodes[0].start_child);
  else
    {
      fprintf(output, "ratio=0.5;\n");
      float current_cost = memory[start_costs + 0].fvalue;
      float delta = current_cost / nodes[0].cost;
      red = (unsigned char)((delta/2.0) * 255);
      green = 255 - red;
      fprintf(output, "Node0 [label=\"\" fixedsize=true width=0.0 height=0.0 color=\"#%02x%02x00\"];\n", red, green);
    }
  writeDOTNode(output, 0, start_scene, memory, verbosity, 0, maxDepth);
  fprintf(output, "}\n");
  fclose(output);
}

// assumes the node itself has already been written, 
// actually writes children nodes and the links to them
void BVH::writeDOTNode(FILE *output, int nodeID, int start_scene, FourByte *memory, int verbosity, int depth, int maxDepth)
{
  if(depth == maxDepth)
    return;
  BVHNode node = loadNode(nodeID, start_scene, memory);

  if(node.num_children >= 0)
    return;
  BVHNode left = loadNode(node.start_child, start_scene, memory);
  BVHNode right = loadNode(node.start_child+1, start_scene, memory);

  char leftClass = 'L', rightClass = 'L';
  if(left.num_children < 0)
    leftClass = 'I';
  if(right.num_children < 0)
    rightClass = 'I';
  
  float current_cost, delta;
  unsigned char red, green; 
  if(verbosity)
    {
      fprintf(output, "Node%d [label=\"%d, %d\"];\n", node.start_child, node.start_child, left.start_child);
      fprintf(output, "Node%d [label=\"%d, %d\"];\n", node.start_child + 1, node.start_child + 1, right.start_child);
      fprintf(output, "Node%d -- Node%d;\n", nodeID, node.start_child);
      fprintf(output, "Node%d -- Node%d;\n", nodeID, node.start_child+1);
    }
  else
    {
      current_cost = memory[start_costs + node.start_child].fvalue;
      delta = current_cost / nodes[node.start_child].cost; // compare to the original tree
      red = (unsigned char)((delta/2.0) * 255);
      green = 255 - red;
      fprintf(output, "Node%d [label=\"\" fixedsize=true width=0.0 height=0.0 color=\"#%02x%02x00\"];\n", node.start_child, red, green);
      fprintf(output, "Node%d -- Node%d [color=\"#%02x%02x00\" weight=100];\n", nodeID, node.start_child, red, green);

      current_cost = memory[start_costs + node.start_child+1].fvalue;
      delta = current_cost / nodes[node.start_child + 1].cost; // compare to the original tree
      red = (unsigned char)((delta/2.0) * 255);
      green = 255 - red;
      fprintf(output, "Node%d [label=\"\" fixedsize=true width=0.0 height=0.0 color=\"#%02x%02x00\"];\n", node.start_child+1, red, green);
      fprintf(output, "Node%d -- Node%d [color=\"#%02x%02x00\" weight=100];\n", nodeID, node.start_child+1, red, green);
    }
  
  writeDOTNode(output, node.start_child, start_scene, memory, verbosity, depth + 1, maxDepth);
  writeDOTNode(output, node.start_child+1, start_scene, memory, verbosity, depth + 1, maxDepth);
}

Triangle BVH::loadTriangle(int tri_addr, FourByte *memory)
{
  Triangle t;
  for(int i=0; i < 3; i++)
    {
      t.p0[i] = memory[tri_addr + i + 0].fvalue;
      t.p1[i] = memory[tri_addr + i + 3].fvalue;
      t.p2[i] = memory[tri_addr + i + 6].fvalue;
    }
  return t;
}

int max(int a, int b)
{
  return a > b ? a : b;
}

int BVH::maxDepth(int start_scene, FourByte *memory, int nodeID)
{
  BVHNode n = loadNode(nodeID, start_scene, memory);
  if(n.num_children > 0) // leaf node
    return 1;
  return 1 + max(maxDepth(start_scene, memory, n.start_child), maxDepth(start_scene, memory, n.start_child + 1));
}

void BVH::verifyTree(int start_scene, FourByte *memory, int nodeID)
{
  if(nodeID < 0 || nodeID >= num_nodes)
    printf("Invalid nodeID: %d\n", nodeID);
  if(memory[start_rotated_flags + nodeID].ivalue != 0)
    printf("Node %d, is_rotated flag should be off\n", nodeID);
  BVHNode n = loadNode(nodeID, start_scene, memory);
  float mins[3] = {1000000.0f, 1000000.0f, 1000000.0f};
  float maxs[3] = {-1000000.0f, -1000000.0f, -1000000.0f};
  bool isleaf = (memory[start_scene + (nodeID * 8) + 6].ivalue > 0 );
  if(isleaf)
    {
      Triangle t = loadTriangle(memory[start_scene + (nodeID * 8) + 7].ivalue,
				memory);
      ExtendBoundByTriangle(mins, maxs, t);
    }
  else
    {
      BVHNode left = loadNode(memory[start_scene + (nodeID * 8) + 7].ivalue, start_scene, memory);
      BVHNode right = loadNode(memory[start_scene + (nodeID * 8) + 7].ivalue + 1, start_scene, memory);
      ExtendBoundByBox(mins, maxs, left.box_min, left.box_max);
      ExtendBoundByBox(mins, maxs, right.box_min, right.box_max);
    }
  
  for(int i=0; i < 3; i++)
    {
      if((n.box_max[i] < maxs[i]) || (n.box_min[i] > mins[i]))
	{
	  if(isleaf)
	    printf("invalid leaf node: ");
	  else
	    printf("invalid inner node: ");
	  printf("%d\tchild bounds = [%f %f %f][%f %f %f], parent bounds = [%f %f %f][%f %f %f]\n",
		 nodeID, mins[0], mins[1], mins[2], maxs[0], maxs[1], maxs[2],
		 n.box_min[0], n.box_min[1], n.box_min[2], n.box_max[0], n.box_max[1], n.box_max[2]);
	  break;
	}
    }
  if(isleaf) 
    return;
  verifyTree(start_scene, memory, memory[start_scene + (nodeID * 8) + 7].ivalue);
  verifyTree(start_scene, memory, memory[start_scene + (nodeID * 8) + 7].ivalue + 1);
  //if(nodeID == 0)
  //verifyNodeCounts(start_scene, memory, nodeID);
}

int BVH::verifyNodeCounts(int start_scene, FourByte *memory, int nodeID)
{
  BVHNode n = loadNode(nodeID, start_scene, memory);
  if(n.num_children > 0) // leaf
    {
      if(n.total_sub_nodes != 1)
	printf("Invalid leaf node: %d, subtree size = %d\n", nodeID, n.total_sub_nodes);
      return 1;
    }
  int left_size = verifyNodeCounts(start_scene, memory, n.start_child);
  int right_size = verifyNodeCounts(start_scene, memory, n.start_child + 1);

  if(n.total_sub_nodes != (left_size + right_size + 1))
    printf("Invalid inner node: %d, subtree size = %d, should be %d\n", nodeID, n.total_sub_nodes, left_size + right_size + 1);

  return left_size + right_size + 1;
}

BVHNode BVH::loadNode(int nodeID, int start_scene, FourByte *memory)
{
  BVHNode retVal;
  int node_addr = start_scene + nodeID * 8;
  for(int i=0; i < 3; i++)
    {
      retVal.box_min[i] = memory[node_addr + i].fvalue;
      retVal.box_max[i] = memory[node_addr + i+3].fvalue;
    }
  retVal.num_children = memory[node_addr + 6].ivalue;  
  retVal.start_child = memory[node_addr + 7].ivalue;
  retVal.total_sub_nodes = memory[start_subtree_sizes + nodeID].ivalue;
  retVal.cost = memory[start_costs + nodeID].fvalue;
  
  return retVal;
}


float BVH::oldComputeNodeCost(int nodeID)
{
  BVHNode *node = &nodes[nodeID];
  if(node->num_children >= 0) // leaf node
    return node->num_children * BVH_C_isec;
  else
    {
      const int leftID = node->start_child;
      const int rightID = leftID + 1;
      float leftCost = oldComputeNodeCost(leftID);
      float rightCost = oldComputeNodeCost(rightID);
      float area = BoxArea(node->box_min, node->box_max);
      float leftArea = BoxArea(nodes[leftID].box_min, nodes[leftID].box_max);
      float rightArea = BoxArea(nodes[rightID].box_min, nodes[rightID].box_max);
      node->cost = BVH_C_trav + (leftArea * leftCost + 
				 rightArea * rightCost) / area;
      return node->cost;
      // TODO: this divide cancels out the multiply by areas on the next level up
      // can be taken out, but this isn't part of the simulation, 
      // so doesn't really matter
    }
}

float BVH::computeNodeCost(BVHNode node, int start_scene, FourByte *memory)
{
  if(node.num_children > 0)
    return node.num_children * BVH_C_isec;

  BVHNode left = loadNode(node.start_child, start_scene, memory);
  BVHNode right = loadNode(node.start_child+1, start_scene, memory);

  float leftCost = computeNodeCost(left, start_scene, memory);
  float rightCost = computeNodeCost(right, start_scene, memory);

  float area = BoxArea(node.box_min, node.box_max);

  float leftArea = BoxArea(left.box_min, left.box_max);
  float rightArea = BoxArea(right.box_min, right.box_max);
  return BVH_C_trav + (leftArea * leftCost + 
		       rightArea * rightCost) / area;
  
  // TODO: this divide cancels out the multiply by areas on the next level up
  // can be taken out, but this isn't part of the simulation, 
  // so doesn't really matter
  
}


float BVH::computeSAHCost(int start_scene, FourByte *memory)
{
  BVHNode root = loadNode(0, start_scene, memory);
  float retVal = computeNodeCost(root, start_scene, memory);
  return retVal;
}

int partitionObjectMedian(int tri_begin, int tri_end, int bestAxis) {
  int num_tris = tri_end - tri_begin;
  if (num_tris <= 4) {
    bestAxis = -1;
    return -1;
  } else {
    bestAxis = 0;
    return tri_begin + num_tris/2;
  }

}

void BVH::rebuild(FourByte *memory)
{


  // delete and remove the old triangles
  for(size_t i = 0; i < triangles->size(); i++)
    delete triangles->at(i);
    
  inorder_tris.clear();
  triangles->clear();

  // get the new triangles form the simulator's memory
  for(int i = 0; i < num_tris; i++)
    {
      Triangle t = loadTriangle(start_tris + (i*11), memory);
      triangles->push_back(new Triangle(t));
    }
  
  int nextFree = 1;
  num_nodes = 0;
  build(0, 0, triangles->size(), nextFree, 0);
  updateBounds(0);
  computeSubtreeSize(0);
  oldComputeNodeCost(0);

  int memory_position = start_nodes;
  LoadNodes(memory_position, INT_MAX, memory, start_nodes);
  LoadTriangles(memory_position, INT_MAX, memory);
}

void BVH::build(int nodeID, int tri_begin, int tri_end,
                int& nextFree, int depth) {
  if (depth == 0) {
    printf("Starting BVH build.\n");
  }

  assert (tri_end > tri_begin);

  BVHNode& node = nodes[nodeID];

  int num_tris = tri_end - tri_begin;
  num_nodes++;
  int bestAxis = -1;
  //int split = partitionObjectMedian(tri_begin, tri_end, bestAxis);
  int split = partitionSAH(tri_begin, tri_end, bestAxis);


  if (bestAxis == -1) {
    // make leaf
    int first_inorder = int(inorder_tris.size());
    for (int i = 0; i < num_tris; i++) {
      inorder_tris.push_back(triangles->at(tri_begin + i));
    }
    node.makeLeaf(first_inorder, num_tris);
  }
  else {
    //int split = tri_begin + num_tris/2;
    // make internal node
    node.makeInternal(nextFree);
    nextFree += 2;

    build(node.start_child + 0,tri_begin,split,nextFree,depth+1);
    build(node.start_child + 1,split,tri_end,nextFree,depth+1);
  }

  if (depth > 31) printf("Uh oh, BVH depth just hit %d\n", depth);
  
  if (depth == 0) {
    for(size_t i=0; i < inorder_tris.size(); i++)      
      tri_orders.push_back(inorder_tris[i]->object_id); 
    printf("BVH build complete with %d nodes.\n", num_nodes);
  }
}

void ResetBound(float min[3], float max[3]) {
  for (int i = 0; i < 3; i++) {
    min[i] = FLT_MAX;
    max[i] = -FLT_MAX;
  }
}

void ExtendBoundByBox(float cur_min[3], float cur_max[3],
                      float box_min[3], float box_max[3]) {
  for (int i = 0; i < 3; i++) {
    if (box_min[i] < cur_min[i])
      cur_min[i] = box_min[i];
    if (box_max[i] > cur_max[i])
      cur_max[i] = box_max[i];
  }
}

void ExtendBoundByTriangle(float cur_min[3], float cur_max[3],
                           const Triangle& tri) {
// do we treat p0 and p1 as storing edges?
#ifdef TRIANGLES_STORE_EDGES
  for (int i = 0; i < 3; i++) {
	const float tmp0 = tri.p0[i]+tri.p1[i];
    if (tmp0 < cur_min[i]) {
      cur_min[i] = tmp0;
    }
    if (tmp0 > cur_max[i]) {
      cur_max[i] = tmp0;
    }

	if (tri.p1[i] < cur_min[i]) {
      cur_min[i] = tri.p1[i];
    }
    if (tri.p1[i] > cur_max[i]) {
      cur_max[i] = tri.p1[i];
    }

	const float tmp2 = tri.p2[i]+tri.p1[i];
	if (tmp2 < cur_min[i]) {
      cur_min[i] =tmp2;
    }
    if (tmp2 > cur_max[i]) {
      cur_max[i] = tmp2;
    }
  }
#else
  for (int i = 0; i < 3; i++) {
    if (tri.p0[i] < cur_min[i]) {
      cur_min[i] = tri.p0[i];
    }
    if (tri.p1[i] < cur_min[i]) {
      cur_min[i] = tri.p1[i];
    }
    if (tri.p2[i] < cur_min[i]) {
      cur_min[i] = tri.p2[i];
    }

    if (tri.p0[i] > cur_max[i]) {
      cur_max[i] = tri.p0[i];
    }
    if (tri.p1[i] > cur_max[i]) {
      cur_max[i] = tri.p1[i];
    }
    if (tri.p2[i] > cur_max[i]) {
      cur_max[i] = tri.p2[i];
    }
  }
#endif
}


void BVH::updateBounds(int ID) {
  BVHNode& node = nodes[ID];
  if (node.isLeaf()) {
    ResetBound(node.box_min, node.box_max);
    for (int i = 0; i < node.num_children; i++) {
      int child_id = node.start_child + i;
      ExtendBoundByTriangle(node.box_min,
                            node.box_max,
                            *inorder_tris[child_id]);
    }
  } else {
    int left_son = node.start_child;
    int right_son = left_son + 1;
    updateBounds(left_son);
    updateBounds(right_son);

    ResetBound(node.box_min, node.box_max);
    BVHNode& left_node = nodes[left_son];
    BVHNode& right_node = nodes[right_son];

    ExtendBoundByBox(node.box_min, node.box_max,
                     left_node.box_min, left_node.box_max);
    ExtendBoundByBox(node.box_min, node.box_max,
                     right_node.box_min, right_node.box_max);
  }
}

int BVH::partitionSAH(int objBegin, int objEnd, int& output_axis)
{
    int num_objects = objEnd - objBegin;
    if ( num_objects == 1 )
    {
        output_axis = -1;
        return -1;
    }

    BVHCostEval best_cost;
    best_cost.event = 0;
#ifdef TREE_ROTATIONS
    best_cost.cost = FLT_MAX;
#else
    best_cost.cost = BVH_C_isec * num_objects;
#endif
    best_cost.axis = -1;

    for ( int axis = 0; axis < 3; axis++ )
    {
        BVHCostEval new_cost;
        if ( buildEvents(objBegin,objEnd,axis,new_cost) )
        {
            if ( new_cost.cost < best_cost.cost )
            {
                best_cost = new_cost;
            }
        }
    }

    output_axis = best_cost.axis;
    if ( output_axis != -1 )
    {
        // build the events and sort them
        std::vector<BVHSAHEvent> events;
        for ( int i = objBegin; i < objEnd; i++ )
        {
          float tri_box_min[3];
          float tri_box_max[3];
          ResetBound(tri_box_min, tri_box_max);
          ExtendBoundByTriangle(tri_box_min, tri_box_max,
                                *triangles->at(i));

          BVHSAHEvent new_event;
          new_event.position = .5f*(tri_box_min[output_axis] +
                                    tri_box_max[output_axis]);
          new_event.obj_id   = i;
          events.push_back(new_event);
        }
        std::sort(events.begin(), events.end(), CompareBVHSAHEvent());
        std::vector<Triangle*> copied_tris;
        for (size_t i = 0; i < events.size(); i++) {
          copied_tris.push_back(triangles->at(events.at(i).obj_id));
        }

        for (int i = objBegin; i < objEnd; i++) {
          triangles->at(i) = copied_tris.at(i-objBegin);
        }

        int result = objBegin + best_cost.event;

        if ( result == objBegin || result == objEnd )
        {
            if ( num_objects < 8 )
            {
                output_axis = -1;
                return 0;
            }
            result = objBegin + num_objects/2;
            return result;
        }
        else
            return result;
    }
    else
    {
        return 0; // making a leaf anyway
    }

}

bool BVH::buildEvents(int first,
                      int last,
                      int axis,
                      BVHCostEval& best_eval)
{
  float overall_box_min[3];
  float overall_box_max[3];
  ResetBound(overall_box_min, overall_box_max);

  std::vector<BVHSAHEvent> events;
  for ( int i = first; i < last; i++ )
  {
    BVHSAHEvent new_event;
    float tri_box_min[3];
    float tri_box_max[3];
    ResetBound(tri_box_min, tri_box_max);
    ExtendBoundByTriangle(tri_box_min, tri_box_max, *triangles->at(i));

    new_event.position = .5f * (tri_box_min[axis] + tri_box_max[axis]);
    new_event.obj_id   = i;
    events.push_back(new_event);

    ExtendBoundByBox(overall_box_min, overall_box_max,
                     tri_box_min, tri_box_max);
  }

  std::sort(events.begin(),events.end(),CompareBVHSAHEvent());

  int num_events = int(events.size());
  float left_box_min[3];
  float left_box_max[3];
  ResetBound(left_box_min, left_box_max);

  int num_left = 0;
  int num_right = num_events;

    for ( size_t i = 0; i < events.size(); i++ )
    {
        events[i].num_left = num_left;
        events[i].num_right = num_right;
        events[i].left_area = BoxArea(left_box_min, left_box_max);

        float tri_box_min[3];
        float tri_box_max[3];
        ResetBound(tri_box_min, tri_box_max);
        ExtendBoundByTriangle(tri_box_min, tri_box_max,
                              *triangles->at(events[i].obj_id));
        ExtendBoundByBox(left_box_min, left_box_max,
                         tri_box_min, tri_box_max);

        num_left++;
        num_right--;
    }

    float right_box_min[3];
    float right_box_max[3];
    ResetBound(right_box_min, right_box_max);

    best_eval.cost = FLT_MAX;
    best_eval.event = -1;

    for ( int i = num_events - 1; i >= 0; i-- )
    {
        float tri_box_min[3];
        float tri_box_max[3];
        ResetBound(tri_box_min, tri_box_max);
        ExtendBoundByTriangle(tri_box_min, tri_box_max,
                              *triangles->at(events[i].obj_id));
        ExtendBoundByBox(right_box_min, right_box_max,
                         tri_box_min, tri_box_max);

        if ( events[i].num_left > 0 && events[i].num_right > 0 )
        {
            events[i].right_area = BoxArea(right_box_min,
                                           right_box_max);

            float this_cost = (events[i].num_left * events[i].left_area +
                               events[i].num_right * events[i].right_area);
            this_cost /= BoxArea(overall_box_min,
                                 overall_box_max);
            this_cost *= BVH_C_isec;
            this_cost += BVH_C_trav;

            events[i].cost = this_cost;
            if ( this_cost < best_eval.cost )
            {
                best_eval.cost       = this_cost;
                best_eval.position   = events[i].position;
                best_eval.axis       = axis;
                best_eval.event      = i;
                best_eval.num_left   = events[i].num_left;
                best_eval.num_right  = events[i].num_right;

            }
        }
    }
    return best_eval.event != -1;
}
